@@ -45,7 +45,7 @@ from collections import OrderedDict
 from osgeo import osr
 from osgeo import ogr
 import shapely.wkt
-from shapely.geometry import Point
+from shapely.geometry import Polygon
 from geopandas import GeoSeries
 from geopandas import GeoDataFrame
 from geopandas.base import is_geometry_type
@@ -54,13 +54,18 @@ from veranda.io.geotiff import GeoTiffFile
 from veranda.io.netcdf import NcFile
 from veranda.io.timestack import GeoTiffRasterTimeStack
 from veranda.io.timestack import NcRasterTimeStack
+from geospade.tools import rasterise_polygon
+from geospade.raster import RasterGeometry
+from geospade.raster import SpatialRef
+from geospade.transform import xy2ij
+from geospade.transform import ij2xy
 
 # load yeoda's utils module
 from yeoda.utils import get_file_type
 from yeoda.utils import any_geom2ogr_geom
-from yeoda.utils import xy2ij
-from yeoda.utils import ij2xy
-from yeoda.utils import boundary
+from yeoda.utils import to_list
+from yeoda.utils import swap_axis
+from yeoda.utils import get_polygon_envelope
 
 # load classes from yeoda's error module
 from yeoda.errors import IOClassNotFound
@@ -71,8 +76,6 @@ from yeoda.errors import DimensionUnkown
 from yeoda.errors import LoadingDataError
 from yeoda.errors import SpatialInconsistencyError
 
-# TODO: resolve geometry/tile columns and simplify queries
-
 
 def _check_inventory(f):
     """
@@ -101,7 +104,6 @@ def _check_inventory(f):
     return f_wrapper
 
 
-# TODO: maybe use pandas isequal after function call to set this flag
 def _set_status(status):
     """
     Decorator for `EODataCube` functions to set internal flag defining if a process changes the data cube structure or
@@ -134,9 +136,10 @@ def _set_status(status):
 class EODataCube:
     """
     A file(name) based data cube for preferably gridded and well-structured EO data.
+
     """
 
-    def __init__(self, filepaths=None, grid=None, smart_filename_class=None, dimensions=None, inventory=None,
+    def __init__(self, filepaths=None, grid=None, filename_class=None, dimensions=None, inventory=None,
                  io_map=None, sdim_name="tile", tdim_name="time"):
         """
         Constructor of the `EODataCube` class.
@@ -147,8 +150,8 @@ class EODataCube:
             List of file paths.
         grid : pytileproj.base.TiledProjection, optional
             Tiled projection/grid object/class (e.g. `Equi7Grid`, `LatLonGrid`).
-        smart_filename_class : geopathfinder.file_naming.SmartFilename, optional
-            `SmartFilename` class to handle the interpretation of filenames.
+        filename_class : geopathfinder.file_naming.SmartFilename, optional
+            `SmartFilename` class to handle the interpretation of file names.
         dimensions : list of str, optional
             List of filename parts to use as dimensions. The strings have to match with the keys of the `SmartFilename`
             fields definition.
@@ -169,6 +172,7 @@ class EODataCube:
         self._ds = None  # data set pointer
         self.status = None
         self.tdim_name = tdim_name
+        self._filename_class = filename_class
 
         # initialise IO classes responsible for reading and writing
         if io_map is not None:
@@ -181,15 +185,15 @@ class EODataCube:
         if inventory is not None:
             self.inventory = inventory
         else:
-            self.__inventory_from_filepaths(filepaths, dimensions=dimensions,
-                                            smart_filename_class=smart_filename_class)
+            self.__inventory_from_filepaths(filepaths, dimensions=dimensions)
 
         self.grid = None
         if grid:
             self.grid = grid
             self.sdim_name = sdim_name
         elif (self.inventory is not None) and ('geometry' not in self.inventory.keys()):
-            geometries = [self.__geometry_from_file(filepath) for filepath in self.filepaths]
+            geometries = [self.__raster_geom_from_file(filepath).boundary_shapely
+                          for filepath in self.filepaths]
             self.sdim_name = sdim_name if sdim_name in self.dimensions else "geometry"
             self.add_dimension('geometry', geometries, inplace=True)
         else:
@@ -223,16 +227,55 @@ class EODataCube:
     @property
     def boundary(self):
         """
-        shapely.geometry : Shapely polygon representing the boundary of the data cube or `None` if no files are
+        ogr.geometry : OGR polygon representing the boundary/envelope of the data cube or `None` if no files are
         contained in the data cube.
+
         """
 
         self.__check_spatial_consistency()
+        boundary = None
         if self.filepaths is not None:
             filepath = self.filepaths[0]
-            return self.__geometry_from_file(filepath)
-        else:
-            return None
+            raster_geom = self.__raster_geom_from_file(filepath)
+            boundary = raster_geom.boundary_ogr
+            boundary = swap_axis(boundary)  # ensure lon-lat order
+
+        return boundary
+
+    @property
+    def raster_geometry(self):
+        """
+        geospade.raster.RasterGeometry :
+            Raster geometry representing the geometric properties of the given file. Note
+            that the raster geometry is extracted from the first file, so be sure that the
+            datacube only holds files from the same tile of the grid.
+
+        """
+        self.__check_spatial_consistency()
+        raster_geom = None
+        if not self.inventory.empty:
+            raster_geom = self.__raster_geom_from_file(self.filepaths[0])
+
+        return raster_geom
+
+    @property
+    def coordinate_boundary(self):
+        """
+        ogr.geometry : OGR polygon representing the coordinate boundary of the data cube or `None` if no files are
+        contained in the data cube.
+
+        """
+
+        self.__check_spatial_consistency()
+        coord_boundary = None
+        if self.filepaths is not None:
+            filepath = self.filepaths[0]
+            raster_geom = self.__raster_geom_from_file(filepath)
+            coord_boundary = ogr.CreateGeometryFromWkt(Polygon(raster_geom.coord_corners).wkt)
+            coord_boundary.AssignSpatialReference(raster_geom.sref.osr_sref)
+            coord_boundary = swap_axis(coord_boundary)  # ensure lon-lat order
+
+        return coord_boundary
 
     @classmethod
     def from_inventory(cls, inventory, grid=None, **kwargs):
@@ -284,7 +327,7 @@ class EODataCube:
 
         inventory = copy.deepcopy(self.inventory)
         inventory = inventory.rename(columns=dimensions_map)
-        return self.__assign_inventory(inventory, inplace=inplace)
+        return self._assign_inventory(inventory, inplace=inplace)
 
     @_check_inventory
     def add_dimension(self, name, values, inplace=False):
@@ -313,7 +356,7 @@ class EODataCube:
         else:
             ds = pd.Series(values, index=self.inventory.index)
         inventory = self.inventory.assign(**{name: ds})
-        return self.__assign_inventory(inventory, inplace=inplace)
+        return self._assign_inventory(inventory, inplace=inplace)
 
     @_set_status('changed')
     @_check_inventory
@@ -345,7 +388,7 @@ class EODataCube:
             file_filter = lambda x: re.search(pattern, x) is not None
         idx_filter = [file_filter(filepath) for filepath in self.filepaths]
         inventory = self.inventory[idx_filter]
-        return self.__assign_inventory(inventory, inplace=inplace)
+        return self._assign_inventory(inventory, inplace=inplace)
 
     @_set_status('changed')
     @_check_inventory
@@ -387,7 +430,7 @@ class EODataCube:
             bool_filter.append(select)
 
         inventory = self.inventory[bool_filter]
-        return self.__assign_inventory(inventory, inplace=inplace)
+        return self._assign_inventory(inventory, inplace=inplace)
 
     @_set_status('changed')
     @_check_inventory
@@ -414,7 +457,7 @@ class EODataCube:
         inventory = copy.deepcopy(self.inventory)
         inventory_sorted = inventory.sort_values(by=name, ascending=ascending)
 
-        return self.__assign_inventory(inventory=inventory_sorted)
+        return self._assign_inventory(inventory=inventory_sorted)
 
     @_set_status('changed')
     def filter_by_dimension(self, values, expressions=None, name="time", inplace=False):
@@ -470,8 +513,7 @@ class EODataCube:
             `EODataCube` object with a filtered inventory according to the given tile names.
         """
 
-        if not isinstance(tilenames, (tuple, list)):
-            tilenames = [tilenames]
+        tilenames = to_list(tilenames)
 
         if use_grid:
             if self.grid is not None:
@@ -517,6 +559,7 @@ class EODataCube:
             sref_lonlat = osr.SpatialReference()
             sref_lonlat.ImportFromEPSG(4326)
             geom_roi.TransformTo(sref_lonlat)
+            geom_roi = swap_axis(geom_roi)
             ftilenames = self.grid.search_tiles_over_geometry(geom_roi)
             tilenames = [ftilename.split('_')[1] for ftilename in ftilenames]
             return self.filter_spatially_by_tilename(tilenames, inplace=inplace, use_grid=False)
@@ -525,7 +568,7 @@ class EODataCube:
             geom_roi = self.align_geom(geom_roi)
             geom_roi = shapely.wkt.loads(geom_roi.ExportToWkt())
             inventory = self.inventory[self.inventory.intersects(geom_roi)]
-            return self.__assign_inventory(inventory, inplace=inplace)
+            return self._assign_inventory(inventory, inplace=inplace)
         else:
             return self
 
@@ -577,9 +620,7 @@ class EODataCube:
         monthly_eodcs = []
         for yearly_eodc in yearly_eodcs:
             if months is not None:
-                yearly_months = months
-                if not isinstance(months, list):
-                    yearly_months = [yearly_months]
+                yearly_months = to_list(months)
                 # initialise empty dict keeping track of the months
                 timestamps_months = {}
                 for month in yearly_months:
@@ -630,8 +671,7 @@ class EODataCube:
 
         sort = False
         if years:
-            if not isinstance(years, list):
-                years = [years]
+            years = to_list(years)
             # initialise empty dict keeping track of the years
             timestamps_years = {}
             for year in years:
@@ -663,14 +703,14 @@ class EODataCube:
         return self.split_by_dimension(values, expressions, name=self.tdim_name)
 
     @_set_status('stable')
-    def load_by_geom(self, geom, sref=None, band=1, apply_mask=False, dtype="xarray", origin='ul',
+    def load_by_geom(self, geom, sref=None, band=1, apply_mask=True, dtype="xarray", origin='ul',
                      decode_kwargs=None):
         """
         Loads data according to a given geometry.
 
         Parameters
         ----------
-        geom : OGR Geometry or Shapely Geometry or list or tuple, optional
+        geom : OGR Geometry or Shapely Geometry or list or tuple
             A geometry defining the region of interest. If it is of type list/tuple representing the extent
             (i.e. [x_min, y_min, x_max, y_max]), `sref` has to be given to transform the extent into a
             georeferenced polygon.
@@ -680,7 +720,7 @@ class EODataCube:
             Band number or name (default is 1).
         apply_mask : bool, optional
             If true, a numpy mask array with a mask excluding (=1) all pixels outside `geom` (=0) will be created
-            (default is True).
+            (default is False).
         dtype : str
             Data type of the returned array-like structure (default is 'xarray'). It can be:
                 - 'xarray': loads data as an xarray.DataSet
@@ -702,7 +742,7 @@ class EODataCube:
             Data as an array-like object.
         """
 
-        if self.inventory is None:  # no data given
+        if self.inventory is None or self.inventory.empty:  # no data given
             return None
 
         decode_kwargs = {} if decode_kwargs is None else decode_kwargs
@@ -727,29 +767,33 @@ class EODataCube:
         roi_sref = geom_roi.GetSpatialReference()
         if not this_sref.IsSame(roi_sref):
             geom_roi = geometry.transform_geometry(geom_roi, this_sref)
+            geom_roi = swap_axis(geom_roi)
 
         # clip region of interest to tile boundary
-        boundary_ogr = ogr.CreateGeometryFromWkt(self.boundary.wkt)
-        geom_roi = geom_roi.Intersection(boundary_ogr)
-        if geom_roi.ExportToWkt() == 'GEOMETRYCOLLECTION EMPTY':
+        geom_roi = geom_roi.Intersection(self.coordinate_boundary)
+        if geom_roi.IsEmpty():
             raise Exception('The given geometry does not intersect with the tile boundaries.')
         geom_roi.AssignSpatialReference(this_sref)
 
-        extent = geometry.get_geometry_envelope(geom_roi)
+        # remove third dimension from geometry
+        geom_roi.FlattenTo2D()
+
+        # retrieve extent of polygon with respect to the pixel sampling of the grid
+        x_pixel_size = abs(this_gt[1])
+        y_pixel_size = abs(this_gt[5])
+        extent = get_polygon_envelope(shapely.wkt.loads(geom_roi.ExportToWkt()),
+                                      x_pixel_size,
+                                      y_pixel_size)
         inv_traffo_fun = lambda i, j: ij2xy(i, j, this_gt, origin=origin)
-        min_col, min_row = xy2ij(extent[0], extent[3], this_gt)
-        max_col, max_row = xy2ij(extent[2], extent[1], this_gt)
-        col_size = max_col - min_col
-        row_size = max_row - min_row
+        min_col, min_row = [int(coord) for coord in xy2ij(extent[0], extent[3], this_gt)]
+        max_col, max_row = [int(coord) for coord in xy2ij(extent[2], extent[1], this_gt)]
+        max_col, max_row = max_col + 1, max_row + 1 # plus one to still include the maximum indexes
+
         if apply_mask:
-            geom_roi = shapely.wkt.loads(geom_roi.ExportToWkt())
-            data_mask = np.ones((row_size, col_size))
-            for col in range(col_size):
-                for row in range(row_size):
-                    x, y = inv_traffo_fun(min_col + col, min_row + row)
-                    point = Point(x, y)
-                    if point.within(geom_roi) or point.touches(geom_roi):
-                        data_mask[row, col] = 0
+            # pixel size extraction assumes non-rotated data
+            data_mask = np.invert(rasterise_polygon(shapely.wkt.loads(geom_roi.ExportToWkt()),
+                                                    x_pixel_size,
+                                                    y_pixel_size).astype(bool))
 
         file_type = get_file_type(self.filepaths[0])
         xs = None
@@ -758,10 +802,14 @@ class EODataCube:
             if self._ds is None and self.status != "stable":
                 file_ts = {'filenames': list(self.filepaths)}
                 self._ds = GeoTiffRasterTimeStack(file_ts=file_ts, file_band=band)
+            col_size = max_col - min_col
+            row_size = max_row - min_row
             data = self._ds.read_ts(min_col, min_row, col_size=col_size, row_size=row_size)
             if data is None:
                 raise LoadingDataError()
             data = self.decode(data, **decode_kwargs)
+            if len(data.shape) == 2:  # ensure that the data is always forwarded as a 3D array
+                data = data[None, :, :]
             if apply_mask:
                 data = np.ma.array(data, mask=np.stack([data_mask]*data.shape[0], axis=0))
 
@@ -830,15 +878,13 @@ class EODataCube:
             Data as an array-like object.
         """
 
-        if self.inventory is None:  # no data given
+        if self.inventory is None or self.inventory.empty:  # no data given
             return None
 
         decode_kwargs = {} if decode_kwargs is None else decode_kwargs
 
-        if not isinstance(rows, list):
-            rows = [rows]
-        if not isinstance(cols, list):
-            cols = [cols]
+        rows = to_list(rows)
+        cols = to_list(cols)
 
         if self.grid:
             if self.sdim_name not in self.dimensions:
@@ -871,7 +917,10 @@ class EODataCube:
                 data_i = self._ds.read_ts(col, row, col_size=col_size, row_size=row_size)
                 if data_i is None:
                     raise LoadingDataError()
-                data.append(self.decode(data_i, **decode_kwargs))
+                data_i = self.decode(data_i, **decode_kwargs)
+                if len(data_i.shape) == 2:  # ensure that the data is always forwarded as a 3D array
+                    data_i = data_i[None, :, :]
+                data.append(data_i)
                 if row_size != 1 and col_size != 1:
                     max_col = col + col_size
                     max_row = row + row_size
@@ -939,15 +988,13 @@ class EODataCube:
             Data as an array-like object.
         """
 
-        if self.inventory is None:  # no data given
+        if self.inventory is None or self.inventory.empty:  # no data given
             return None
 
         decode_kwargs = {} if decode_kwargs is None else decode_kwargs
 
-        if not isinstance(xs, list):
-            xs = [xs]
-        if not isinstance(ys, list):
-            ys = [ys]
+        xs = to_list(xs)
+        ys = to_list(ys)
 
         if self.grid is not None:
             if self.sdim_name not in self.dimensions:
@@ -970,7 +1017,11 @@ class EODataCube:
 
             if sref is not None:
                 x, y = geometry.uv2xy(x, y, sref, this_sref)
-            col, row = xy2ij(x, y, this_gt)
+            col, row = [int(coord) for coord in xy2ij(x, y, this_gt)]
+            # check if coordinate is within datacube
+            raster_geom = self.__raster_geom_from_file(self.filepaths[0])
+            if (col < 0) or (row < 0)  or (col >= raster_geom.n_cols) or (row >= raster_geom.n_rows):
+                raise Exception('The given coordinate does not intersect with the tile boundaries.')
             # replace old coordinates with transformed coordinates related to the users definition
             x_t, y_t = ij2xy(col, row, this_gt, origin=origin)
             xs[i] = x_t
@@ -986,6 +1037,8 @@ class EODataCube:
                     raise LoadingDataError()
 
                 data_i = self.decode(data_i, **decode_kwargs)
+                if len(data_i.shape) == 2:  # ensure that the data is always forwarded as a 3D array
+                    data_i = data_i[None, :, :]
             elif file_type == "NetCDF":
                 if self._ds is None and self.status != "stable":
                     file_ts = pd.DataFrame({'filenames': list(self.filepaths)})
@@ -1068,7 +1121,7 @@ class EODataCube:
         """
 
         dc_intersected = intersect_datacubes([self, dc_other], on_dimension=on_dimension)
-        return self.__assign_inventory(dc_intersected.inventory, inplace=inplace)
+        return self._assign_inventory(dc_intersected.inventory, inplace=inplace)
 
     @_set_status('changed')
     @_check_inventory
@@ -1095,7 +1148,7 @@ class EODataCube:
         """
 
         dc_united = unite_datacubes([self, dc_other])
-        return self.__assign_inventory(dc_united.inventory, inplace=inplace)
+        return self._assign_inventory(dc_united.inventory, inplace=inplace)
 
     @_set_status('changed')
     def align_dimension(self, dc_other, name, inplace=False):
@@ -1130,7 +1183,7 @@ class EODataCube:
         idxs = idxs[idxs != -1]
         if len(idxs) > 0:
             inventory = self.inventory.iloc[idxs].reset_index(drop=True)
-            return self.__assign_inventory(inventory, inplace=inplace)
+            return self._assign_inventory(inventory, inplace=inplace)
         else:
             print('No common dimension values found. Original data cube is returned.')
             return self
@@ -1169,6 +1222,7 @@ class EODataCube:
         geom = any_geom2ogr_geom(geom, osr_sref=sref)
         this_sref, _ = self.__get_georef()
         geom = geometry.transform_geometry(geom, this_sref)
+        geom = swap_axis(geom)
 
         return geom
 
@@ -1213,10 +1267,8 @@ class EODataCube:
                 for i, entry in enumerate(data):
                     x = xs[i]
                     y = ys[i]
-                    if not isinstance(x, list):
-                        x = [x]
-                    if not isinstance(y, list):
-                        y = [y]
+                    x = to_list(x)
+                    y = to_list(y)
                     xr_ar = xr.DataArray(entry, coords={self.tdim_name: timestamps, 'y': y, 'x': x},
                                          dims=[self.tdim_name, 'y', 'x'])
                     ds.append(xr.Dataset(data_vars={str(band): xr_ar}))
@@ -1255,8 +1307,9 @@ class EODataCube:
             else:
                 raise DataTypeUnknown(type(data), dtype)
         elif dtype == "dataframe":
+            dimensions = ["time", "y", "x"]
             xr_ds = self.__convert_dtype(data, 'xarray', xs=xs, ys=ys, band=band)
-            converted_data = xr_ds.to_dataframe()
+            converted_data = xr_ds.to_dataframe().reset_index().sort_values(dimensions).set_index(dimensions)
         else:
             raise DataTypeUnknown(type(data), dtype)
 
@@ -1313,9 +1366,9 @@ class EODataCube:
         else:
             return self.io_map[file_type]
 
-    def __geometry_from_file(self, filepath):
+    def __raster_geom_from_file(self, filepath):
         """
-        Retrieves boundary geometry from an EO file.
+        Retrieves a raster geometry from an EO file.
 
         Parameters
         ----------
@@ -1324,18 +1377,20 @@ class EODataCube:
 
         Returns
         -------
-        shapely.geometry
-            Shapely polygon representing the boundary of the file or `None` if the file can not be identified.
+        geospade.raster.RasterGeometry
+            Raster geometry representing the geometric properties of the given file.
+
         """
 
         file_type = get_file_type(filepath)
         io_class = self.__io_class(file_type)
         io_instance = io_class(filepath, mode='r')
-        boundary_geom = boundary(io_instance.geotransform, io_instance.spatialref, io_instance.shape)
+        sref = SpatialRef(io_instance.spatialref)
+        raster_geom = RasterGeometry(io_instance.shape[0], io_instance.shape[1], sref, io_instance.geotransform)
         # close data set
         io_instance.close()
 
-        return shapely.wkt.loads(boundary_geom.ExportToWkt())
+        return raster_geom
 
     def __check_spatial_consistency(self):
         """
@@ -1355,7 +1410,7 @@ class EODataCube:
             if len(uni_vals) > 1:
                 raise SpatialInconsistencyError()
 
-    def __inventory_from_filepaths(self, filepaths, dimensions=None, smart_filename_class=None):
+    def __inventory_from_filepaths(self, filepaths, dimensions=None):
         """
         Creates GeoDataFrame (`inventory`) based on all filepaths.
         Each filepath/filename is translated to a SmartFilename object using a translation function
@@ -1368,8 +1423,7 @@ class EODataCube:
         dimensions : list of str, optional
             List of filename parts to use as dimensions. The strings have to match with the keys of the `SmartFilename`
             fields definition.
-        smart_filename_class : geopathfinder.file_naming.SmartFilename, optional
-            `SmartFilename` class to handle the interpretation of filenames.
+
         """
 
         inventory = OrderedDict()
@@ -1384,7 +1438,7 @@ class EODataCube:
             # get information from filename
             smart_filename = None
             try:
-                smart_filename = smart_filename_class.from_filename(os.path.basename(filepath), convert=True)
+                smart_filename = self._filename_class.from_filename(os.path.basename(filepath), convert=True)
             except:
                 pass
 
@@ -1450,25 +1504,18 @@ class EODataCube:
             If not, the inventory of the data cube is filtered.
         """
 
-        if not isinstance(values, list):
-            values = [values]
+        values = to_list(values)
         n_filters = len(values)
         if expressions is None:  # equal operator is the default comparison operator
             expressions = ["=="] * n_filters
         else:
-            if not isinstance(expressions, list):
-                values = [expressions]
+            expressions = to_list(expressions)
 
         inventory = copy.deepcopy(self.inventory)
         filtered_inventories = []
         for i in range(n_filters):
-            value = values[i]
-            expression = expressions[i]
-            if not isinstance(value, (tuple, list)):
-                value = [value]
-
-            if not isinstance(expression, (tuple, list)):
-                expression = [expression]
+            value = to_list(values[i])
+            expression = to_list(expressions[i])
 
             if (len(value) == 2) and (len(expression) == 2):
                 filter_cmd = "inventory[(inventory[name] {} value[0]) & " \
@@ -1481,14 +1528,14 @@ class EODataCube:
             filtered_inventories.append(eval(filter_cmd))
 
         if split:
-            eodcs = [self.__assign_inventory(filtered_inventory, inplace=False)
+            eodcs = [self._assign_inventory(filtered_inventory, inplace=False)
                      for filtered_inventory in filtered_inventories]
             return eodcs
         else:
             filtered_inventory = pd.concat(filtered_inventories, ignore_index=True)
-            return self.__assign_inventory(filtered_inventory, inplace=inplace)
+            return self._assign_inventory(filtered_inventory, inplace=inplace)
 
-    def __assign_inventory(self, inventory, inplace=True):
+    def _assign_inventory(self, inventory, inplace=True):
         """
         Helper method for either create a new data cube or overwrite the old data cube with the given inventory.
 
@@ -1522,7 +1569,8 @@ class EODataCube:
             return self
         else:
             return self.from_inventory(inventory=inventory, grid=self.grid,
-                                       tdim_name=tdim_name, sdim_name=sdim_name)
+                                       dimensions=self.dimensions, filename_class=self._filename_class,
+                                       io_map=self.io_map, tdim_name=tdim_name, sdim_name=sdim_name)
 
     def __deepcopy__(self, memodict={}):
         """
