@@ -50,8 +50,8 @@ from tests.setup_test_data import dirpath_test
 from geopathfinder.naming_conventions.sgrt_naming import SgrtFilename
 
 # import yeoda data cube
-from yeoda.datacube import EODataCube
-from yeoda.products.preprocessed import SIG0DataCube
+from src.yeoda.datacube import EODataCube
+from src.yeoda.products.preprocessed import SIG0DataCube
 
 
 class LoadingTester(unittest.TestCase):
@@ -80,7 +80,7 @@ class LoadingTester(unittest.TestCase):
         appropriately.
         """
 
-        dc = EODataCube(filepaths=filepaths, smart_filename_class=SgrtFilename,
+        dc = EODataCube(filepaths=filepaths, filename_class=SgrtFilename,
                         dimensions=['time', 'var_name', 'pol', 'tile_name', 'orbit_direction'], sdim_name="tile_name")
 
         dc.filter_by_dimension('VV', name='pol', inplace=True)
@@ -108,18 +108,20 @@ class LoadingCoordsTester(LoadingTester):
         self.lat = 44.
         sref = osr.SpatialReference()
         sref.ImportFromEPSG(4326)
+        sref.SetAxisMappingStrategy(0)  # ensure order is lon-lat
         self.sref = sref
         row = 970
         col = 246
         self.x = 4323250.
         self.y = 1314750.
+        dimensions = ["time", "y", "x"]
 
         self.ref_np_ar = (np.array([[[row + col]*4]]).T + np.arange(0, 4)[:, None, None]).astype(float)
         xr_ar = xr.DataArray(data=da.array(self.ref_np_ar.astype(float)).rechunk((1, 1, 1)),
                              coords={'time': self.timestamps, 'y': [self.y], 'x': [self.x]},
                              dims=['time', 'y', 'x'])
         self.ref_xr_ds = xr.Dataset(data_vars={'1': xr_ar})
-        self.ref_pd_df = self.ref_xr_ds.to_dataframe()
+        self.ref_pd_df = self.ref_xr_ds.to_dataframe().reset_index().sort_values(dimensions).set_index(dimensions)
 
     def test_load_gt2numpy_by_coord(self):
         """ Tests loading of a Numpy array from GeoTIFF files by geographic coordinates. """
@@ -127,11 +129,11 @@ class LoadingCoordsTester(LoadingTester):
         dc = self._create_loadable_dc(self.gt_filepaths)
 
         data = dc.load_by_coords(self.lon, self.lat, sref=self.sref, dtype='numpy')
-        assert (self.ref_np_ar == data).all()
+        assert np.array_equal(self.ref_np_ar, data)
 
         ref_data_list = [self.ref_np_ar, self.ref_np_ar]
         data = dc.load_by_coords([self.lon, self.lon], [self.lat, self.lat], sref=self.sref, dtype='numpy')
-        assert (ref_data_list[0] == data[0]).all() & (ref_data_list[1] == data[1]).all()
+        assert np.array_equal(ref_data_list[0], data[0]) & np.array_equal(ref_data_list[1], data[1])
 
     def test_load_gt2xarray_by_coord(self):
         """ Tests loading of an xarray array from GeoTIFF files by geographic coordinates. """
@@ -167,12 +169,12 @@ class LoadingCoordsTester(LoadingTester):
         dc = self._create_loadable_dc(self.nc_filepaths)
 
         data = dc.load_by_coords(self.lon, self.lat, sref=self.sref, dtype='numpy')
-        assert (self.ref_np_ar == data).all()
+        assert np.array_equal(self.ref_np_ar, data)
         dc.close()
 
         ref_data_list = [self.ref_np_ar, self.ref_np_ar]
         data = dc.load_by_coords([self.lon, self.lon], [self.lat, self.lat], sref=self.sref, dtype='numpy')
-        assert (ref_data_list[0] == data[0]).all() & (ref_data_list[1] == data[1]).all()
+        assert np.array_equal(ref_data_list[0], data[0]) & np.array_equal(ref_data_list[1], data[1])
         dc.close()
 
     def test_load_nc2xarray_by_coord(self):
@@ -207,7 +209,7 @@ class LoadingCoordsTester(LoadingTester):
     def test_load_singlenc2xarray_by_coord(self):
         """ Tests loading of an xarray array from a multidimensional NetCDF file by geographic coordinates. """
 
-        dc = EODataCube(filepaths=[self.nc_filepath], smart_filename_class=SgrtFilename,
+        dc = EODataCube(filepaths=[self.nc_filepath], filename_class=SgrtFilename,
                         dimensions=['time'])
         data = dc.load_by_coords(self.lon, self.lat, sref=self.sref, band='SIG0', dtype='xarray', origin='c')
         assert self.ref_xr_ds.equals(data.rename({'SIG0': '1'}))
@@ -234,13 +236,14 @@ class LoadingPixelsTester(LoadingTester):
         x = 4323250.
         y = 1314750.
         sres = 500.
+        dimensions = ["time", "y", "x"]
 
         self.ref_np_ar = (np.array([[[self.row + self.col]*4]]).T + np.arange(0, 4)[:, None, None]).astype(float)
         xr_ar = xr.DataArray(data=self.ref_np_ar.astype(float),
                              coords={'time': self.timestamps, 'y': [y], 'x': [x]},
                              dims=['time', 'y', 'x'])
         self.ref_xr_ds = xr.Dataset(data_vars={'1': xr_ar})
-        self.ref_pd_df = self.ref_xr_ds.to_dataframe()
+        self.ref_pd_df = self.ref_xr_ds.to_dataframe().reset_index().sort_values(dimensions).set_index(dimensions)
         rows, cols = np.meshgrid(np.arange(self.row, self.row+self.row_size),
                                  np.arange(self.col, self.col+self.col_size), indexing='ij')
         xs = np.arange(x, x + self.col_size * sres, sres)
@@ -252,7 +255,7 @@ class LoadingPixelsTester(LoadingTester):
                              coords={'time': self.timestamps, 'y': ys, 'x': xs},
                              dims=['time', 'y', 'x'])
         self.ref_xr_ds_area = xr.Dataset(data_vars={'1': xr_ar})
-        self.ref_pd_df_area = self.ref_xr_ds_area.to_dataframe()
+        self.ref_pd_df_area = self.ref_xr_ds_area.to_dataframe().reset_index().sort_values(dimensions).set_index(dimensions)
 
     def test_load_gt2numpy_by_pixels(self):
         """ Tests loading of a Numpy array from GeoTIFF files by pixel coordinates. """
@@ -260,10 +263,10 @@ class LoadingPixelsTester(LoadingTester):
         dc = self._create_loadable_dc(self.gt_filepaths)
 
         data = dc.load_by_pixels(self.row, self.col, dtype='numpy')
-        assert (self.ref_np_ar == data).all()
+        assert np.array_equal(self.ref_np_ar, data)
 
         data = dc.load_by_pixels(self.row, self.col, row_size=self.row_size, col_size=self.col_size, dtype='numpy')
-        assert (self.ref_np_ar_area == data).all()
+        assert np.array_equal(self.ref_np_ar_area, data)
 
     def test_load_gt2xarray_by_pixels(self):
         """ Tests loading of an xarray array from GeoTIFF files by pixel coordinates. """
@@ -300,11 +303,11 @@ class LoadingPixelsTester(LoadingTester):
         dc = self._create_loadable_dc(self.nc_filepaths)
 
         data = dc.load_by_pixels(self.row, self.col, dtype='numpy')
-        assert (self.ref_np_ar == data).all()
+        assert np.array_equal(self.ref_np_ar, data)
         dc.close()
 
         data = dc.load_by_pixels(self.row, self.col, row_size=self.row_size, col_size=self.col_size, dtype='numpy')
-        assert (self.ref_np_ar_area == data).all()
+        assert np.array_equal(self.ref_np_ar_area, data)
         dc.close()
 
     def test_load_nc2xarray_by_pixels(self):
@@ -342,7 +345,7 @@ class LoadingPixelsTester(LoadingTester):
     def test_load_singlenc2xarray_by_pixels(self):
         """ Tests loading of an xarray array from a multidimensional NetCDF file by pixel coordinates. """
 
-        dc = EODataCube(filepaths=[self.nc_filepath], smart_filename_class=SgrtFilename,
+        dc = EODataCube(filepaths=[self.nc_filepath], filename_class=SgrtFilename,
                         dimensions=['time'])
         data = dc.load_by_pixels(self.row, self.col, row_size=self.row_size, col_size=self.col_size, band='SIG0',
                                  dtype='xarray', origin='c')
@@ -370,6 +373,7 @@ class LoadingGeomTester(LoadingTester):
         row_size = 10
         col_size = 16
         sres = 500.
+        dimensions = ["time", "y", "x"]
 
         # define bounding boxes
         # general bounding box for reading data
@@ -388,10 +392,10 @@ class LoadingGeomTester(LoadingTester):
         self.partial_outside_bbox = [(x_min, y_min), (x_max, y_max)]
         self.outside_bbox = [(0, 0), (1, 1)]
 
-        rows, cols = np.meshgrid(np.arange(row, row+row_size),
-                                 np.arange(col, col+col_size), indexing='ij')
-        xs = np.arange(x, x + col_size * sres, sres)
-        ys = np.arange(y, y - row_size * sres, -sres)
+        rows, cols = np.meshgrid(np.arange(row, row + row_size + 1),
+                                 np.arange(col, col + col_size + 1), indexing='ij')
+        xs = np.arange(x, x + (col_size + 1) * sres, sres)
+        ys = np.arange(y, y - (row_size + 1) * sres, -sres)
         base_np_ar_2D = rows + cols
         base_np_ar = np.stack([base_np_ar_2D]*4, axis=0)
         self.ref_np_ar_area = (base_np_ar + np.arange(0, 4)[:, None, None]).astype(float)
@@ -399,14 +403,14 @@ class LoadingGeomTester(LoadingTester):
                              coords={'time': self.timestamps, 'y': ys, 'x': xs},
                              dims=['time', 'y', 'x'])
         self.ref_xr_ds_area = xr.Dataset(data_vars={'1': xr_ar})
-        self.ref_pd_df_area = self.ref_xr_ds_area.to_dataframe()
+        self.ref_pd_df_area = self.ref_xr_ds_area.to_dataframe().reset_index().sort_values(dimensions).set_index(dimensions)
 
     def test_load_gt2numpy_by_geom(self):
         """ Tests loading of a Numpy array from GeoTIFF files by a bounding box. """
 
         dc = self._create_loadable_dc(self.gt_filepaths)
         data = dc.load_by_geom(self.bbox, dtype='numpy')
-        assert (self.ref_np_ar_area == data).all()
+        assert np.array_equal(self.ref_np_ar_area, data)
 
     def test_load_gt2xarray_by_geom(self):
         """ Tests loading of an xarray array from GeoTIFF files by a bounding box. """
@@ -429,7 +433,7 @@ class LoadingGeomTester(LoadingTester):
 
         dc = self._create_loadable_dc(self.nc_filepaths)
         data = dc.load_by_geom(self.bbox, dtype='numpy')
-        assert (self.ref_np_ar_area == data).all()
+        assert np.array_equal(self.ref_np_ar_area, data)
         dc.close()
 
     def test_load_nc2xarray_by_geom(self):
@@ -451,7 +455,7 @@ class LoadingGeomTester(LoadingTester):
     def test_load_singlenc2xarray_by_pixels(self):
         """ Tests loading of an xarray array from a multidimensional NetCDF file by a bounding box. """
 
-        dc = EODataCube(filepaths=[self.nc_filepath], smart_filename_class=SgrtFilename,
+        dc = EODataCube(filepaths=[self.nc_filepath], filename_class=SgrtFilename,
                         dimensions=['time'])
         data = dc.load_by_geom(self.bbox, band='SIG0', dtype='xarray', origin='c')
         assert self.ref_xr_ds_area.equals(data.rename({'SIG0': '1'}))
@@ -460,7 +464,8 @@ class LoadingGeomTester(LoadingTester):
     def test_load_geom_larger_than_tile(self):
         """ Tests loading data when the region of interest is larger than the data cube extent. """
 
-        dc = SIG0DataCube(filepaths=self.gt_filepaths, dimensions=['time'], sres=500, sdim_name="tile_name")
+        dc = SIG0DataCube(filepaths=self.gt_filepaths, dimensions=['time'], sres=500, sdim_name="tile_name",
+                          filename_class=SgrtFilename)
         dc.filter_spatially_by_tilename("E042N012T6", inplace=True)
         data = dc.load_by_geom(self.partial_outside_bbox, dtype='numpy')
         assert data.shape == (16, 10, 16)
@@ -468,7 +473,8 @@ class LoadingGeomTester(LoadingTester):
     def test_load_geom_no_intersection(self):
         """ Tests loading data when the region of interest is outside the data cube extent. """
 
-        dc = SIG0DataCube(filepaths=self.gt_filepaths, dimensions=['time'], sres=500, sdim_name="tile_name")
+        dc = SIG0DataCube(filepaths=self.gt_filepaths, dimensions=['time'], sres=500, sdim_name="tile_name",
+                          filename_class=SgrtFilename)
         dc.filter_spatially_by_tilename("E042N012T6", inplace=True)
         try:
             _ = dc.load_by_geom(self.outside_bbox, dtype='numpy')
@@ -476,6 +482,14 @@ class LoadingGeomTester(LoadingTester):
         except:
             assert True
 
+    def test_when_load_geom_and_inventory_empty_then_return_none(self):
+        """ Tests loading data when the data cube is empty. """
+
+        dc = EODataCube(filepaths=[self.nc_filepath], filename_class=SgrtFilename, dimensions=['time'])
+        dc = dc.filter_by_dimension(123, name="time")
+        assert dc.inventory.empty
+        assert dc.load_by_geom(self.outside_bbox, dtype="numpy") is None
+
 
 if __name__ == '__main__':
     unittest.main()
